// FF.cpp : Plugin for MpxManager providing Flat-field corection of measured frames
// by Jan Jakubek,  1.2.2005
// revission:      11.8.2005


// Principle:
// ----------

// During initialization the instance of FFmask class will be created for each Medipix device.
// This object contains information about number of chips and pixels of device and allocates space for
// frame raw data and array of coefficients for Flat-field correction.

// Plugin registers two Menu functions for each device (none specific input or output parameters)
// this functions will allow to:
// 1. choose (using common OpenFileDialog) ASCII file with flat-field coefficients for appropriate device.
// 2. use sum of currently measured frames as a flat-field coeficient

// Plugin registers three general functions to be accessible from other plugins and MpxManager:
// 1. The First function takes three input parameters (Device ID, Frame Index and Frame size),
// gets selected frame raw data from selected device and makes flat-field correction.
// Resulting corrected data (double array of the same size as specified by input parameter)
// will be copied to output parameter.
// 2. Function takes the same set of input parameters but Frame index is ignored. Flat field 
// correction is applied to sum of all frames in the frame buffer. Result will be copied to  
// output parameter (double array of the same size as specified by input parameter)
// 3. Function takes the same set of input parameters but Frame index is ignored. Function copies current 
// FF mask to output parameter (double array of the same size as specified by input parameter)

// Limitations:
//-------------

// Plugin doesn't support hot-swap!
// If set of devices is changed plugin doesn't reflect this change!
// This can lead to use of incorrect flat-field coefficients!

// Functions used from Manager:
//-----------------------------
// mpxCtrlGetFirstMpx
// mpxCtrlGetNextMpx
// mpxCtrlGetMedipixInfo
// mpxCtrlGetFrame32
// mpxCtrlGetFrameDouble
// addMpxMenuItem
// addFuncItem
// logMsg


//=========================================================================================
// Included header files:

#include <windows.h>     // Unfortunately we are under MS windows :-(
#include <stdio.h>       // sprintf, fprintf ... functions are used
#include <commdlg.h>     // Common OpenFileDialog is used
#include <math.h>        // sqrt, fabs


// Each plugin have to include this header file:
//
#include "MpxPluginMgrAPI.h"

#define PLUGIN_NAME     "FF"

//=========================================================================================
// Class FFmask for standard flat-field correction
// - Contains information describing device (name, number of chips, rows, pixels ...)
//   this information is acquired during initialization by function Init.
// - Manages buffer of frame data (32bit per pixel to be able to convert also integral measurements)
//   a source data are copied to this buffer when FF correction of a frame is required (function MakeFFcorr)
// - Manages buffer of flat-field coeficients. Coefficients stored in this array are multiplicative (to avoid division by zero)
//   but they are expected to be loaded from ASCII file as reciprocal values (denominators). They are converted to multiplicative
//   coefficients when they are loaded.

class FFmask
{
public:
    DEVID devID;                            // Mask is valid for this device

    char ifaceName[64];                     // Name of hardware ifaceName ("Muros", "USB", ...)
    char chipboardID[64];                   // Unique name or ID of the chipboardID
    char errorMsg[MPXCTRL_MAX_MSGLENGTH];   // Error message from GetFrame calls
    int numberOfChips;                      // Number of chips
    int numberOfRows;                       // Layout of the chipboardID

    int size;                               // Number of pixels (size of Data and Coef arrays)
    u32 *data;                              // Array of measured data (to be corrected)
    double *mask;                           // Array of FF mask data (before preprocessing)
    double median;                          // Median value of mask array
    double *coef;                           // Array of coeficients (after preprocessing) for flat-field, coefficients are multiplicative !

    double *avrg;                           // Array of average of all frames
    double *stddev;                         // Array of standard deviation of all frames
    int    frames;                          // Number of frames in framebuffer

    char *lastFileName;                     // Name of last used not default filename in LoadCoefMask.
    char defFileName[384];                  // Default file name.
    bool changed;                           // If true array of coefficients was changed and needs to be saved

public:
    FFmask();                               // Default constructor - No initialization
    ~FFmask();                              // Destructor - deletes all allocated buffers, calls SaveCoefMask if needed

public:
    void Delete();                          // Calls SaveCoefMask if needed, Unallocates memory for data and coef arrays
    bool Init(DEVID devID);                 // Acquires device information from MpxManager and allocates memory for data nad coef, calls LoadCoefMask with default filename.
    void SetDefaultMask();                  // Will use default mask (1 in each pixel)
    void PreprocessMask();                  // Makes preprocessing
    bool GetMaskFromFrames();               // Gets mask array from currently measured set of frames (as sum of all)
    bool LoadMask(char *fileName = 0);      // Loads mask array from the file and calls PreprocessMask. If filename is NULL default will be used.
    bool SaveMask(char *fileName = 0);      // Saves mask array to ASCII file. If filename is NULL default name will be used.
    char *GetDefFileName();                 // Returns default filename
    char *GetFileName()                     {return lastFileName ? lastFileName : GetDefFileName();}
    
protected:
    void SetFileName(char *fileName)        {if (lastFileName) free(lastFileName);lastFileName = fileName ? _strdup(fileName): NULL;}

public:
    // MakeFFcorr calls GetFrame32 function from MpxManager. Acquired data are stored in Data array.
    // Uses Coef array to compute corrected Frame. Corrected frame is copied to FFData array. dsize is number of items in FFData array.
    bool MakeFFcorr(FRAMEID frameID);

    // GetAverage and TakeAverageAndSigma call GetFrame32 function from MpxManager in a loop and takes average of all measured frames and their
    // Standard deviation. These arrays are stored in member arrays avrg and stddev. Member frames will be also filled.
    bool TakeAverage();              
    bool TakeAverageAndSigma();              

    // MakeFFcorrOfAverage calls GetFrame32 function from MpxManager in a loop and takes average of all measured frames. Acquired data are stored in Data array.
    // Uses Coef array to compute corrected Frame. Corrected frame is copied to FFData array. dsize is number of items in FFData array.
    bool MakeFFcorrOfAverage(double *FFData, int dsize);

    // Just returns some member arrays:    
    bool GetMask(double *array, int dsize);
    bool GetAverage(double *array, int dsize);
    bool GetSigma(double *array, int dsize);
    
    // MakeSNR computes SNR of each pixel. Signal is diference between average and mask. Noise is StdDev.
    bool MakeSNR(double *array, int dsize);
};


// Class managing dynamic array of FFmask objects for all devices.
// Objects are created and initialized by function Init().
class FFmaskArrayType
{
public:
    unsigned count;                     // Number of FFmask objects for devices
    FFmask *mask;                       // Pointer to allocated array or NULL
public:
    FFmaskArrayType()                   {count = 0;mask = NULL;}    // Default constructor - no FFmask object will be created
    ~FFmaskArrayType()                  {Delete();}                 // Destructor - deletes all FFmask objects
public:
    void Init();                        // Initialization of the array
    void Delete();                      // Deletes all FFmask Objects
public:
    FFmask *GetMask(DEVID devID);       // Searches array for FFmask object with given DEVID
    FFmask *operator[](unsigned i)      {return i < count ? &mask[i]: NULL;}
};


// Forward declarations of functions to be registered by MpxManager
void LoadFFcoefficients(MENUFUNCPAR param);            // Menu function. Loading of flat-field coeficient array from ASCII file.
void GetFFcoefficientsFromFrames(MENUFUNCPAR param);   // Menu function: takes sum of currently measured images and uses it as FF mask

int MakeFFCorrection(INTPTR userData, byte *in, byte *out);             // Function making FF corection of selected frame.
int ShowFFMask(INTPTR userData, byte *in, byte *out);                   // Function will return current FF mask

int GetAverage(INTPTR userData, byte *in, byte *out);
int GetStdDev(INTPTR userData, byte *in, byte *out);

int MakeFFCorrectionOfAverage(INTPTR userData, byte *in, byte *out);    // Function making FF corection of average of all frames.
int EvaluateSNR(INTPTR userData, byte *in, byte *out);                  // Function evaluating SNR for each pixel: Signal is difference between FF mask and average data, noise is StdDev


//=========================================================================================
// Global variables

// Dynamic array of FFmask objects for devices:
// class is just created not initialized
FFmaskArrayType FFmaskArray;


// Description of parameters of MakeFFCorrection function.
// This description is needed to be able to register it by MpxManager:
ExtFunctionInfo MakeFFCorrectionInfo =
{
    MakeFFCorrection,                               // Pointer to function performing an action
    2,                                              // Number of input parameters          = (FRAMEID, INSTID)
    0,                                              // Number of output parameters
    {                                               // List of input parameter descriptions
        {TYPE_U32, 1, "Frame ID"},                  //                                     = first input parameter is ID of device, this device will be source of data to correct
        {TYPE_U32, 1, "Instance ID (not used)"},    //                                     = second input parameter is index of the frame to correct
    },
    {                                               // List of output parameter descriptions
    }, 
    "FF",                                           // Name of the plugin                  = name of this plugin "FF"
    "Flat-field Correction",                        // Name of registered function         = can differ to "C" name: "Flat-Field Correction"
    "Flat-field correction of frame with specified ID",// Description of function purpose     = description of this function
    0,                                              // custom user data
    FUNCFLAG_FILTER                                 // flags                               = flags
};

// Description of parameters of MakeFFCorrectionOfSum function.
// This description is needed to be able to register it by MpxManager:
ExtFunctionInfo MakeFFCorrectionOfAverageInfo =
{
    MakeFFCorrectionOfAverage,                      // Pointer to function performing an action
    2,                                              // Number of input parameters          = (FRAMEID, INSTID)
    0,                                              // Number of output parameters
    {                                               // List of input parameter descriptions
        {TYPE_U32, 1, "Frame ID"},                  //                                     = first input parameter is ID of device, this device will be source of data to correct
        {TYPE_U32, 1, "Instance ID (not used)"},    //                                     = second input parameter is index of the frame to correct
    },
    {                                               // List of output parameter descriptions
    }, 
    "FF",                                           // Name of the plugin                  = name of this plugin "FF"
    "Flat-field Corr. of average",                  // Name of registered function         = can differ to "C" name: "Flat-Field Correction"
    "Flat-field correction of average of all data in frame buffer",// Description of function purpose     = description of this function
    0,                                              // custom user data
    FUNCFLAG_FILTER                                 // flags                               = flags
};

// Description of parameters of ShowFFMaskInfo function.
// This description is needed to be able to register it by MpxManager:
ExtFunctionInfo ShowFFMaskInfo =
{
    ShowFFMask,                                     // Pointer to function performing an action
    3,                                              // Number of input parameters          = (DEVID ID, i32 framenumber, i32 size)
    1,                                              // Number of output parameters         = (double *data)
    {                                               // List of input parameter descriptions
        {TYPE_U32, 1, "DeviceID"},                  //                                     = first input parameter is ID of device, this device will be source of data to correct
        {TYPE_U32, 1, "FrameIndex - dummy"},        //                                     = second input parameter is index of the frame to correct (has no meaning here)
        {TYPE_U32, 1, "PixelCount"}                 //                                     = third input parameter is count of pixels - only for security reasons
    },
    {                                               // List of output parameter descriptions
        {TYPE_DOUBLE,  - 1, "Mask data"}            //                                     = output: double array. Size is given by third input parameter (PixelCount)
    }, 
    "FF",                                           // Name of the plugin                  = name of this plugin "FF"
    "Flat-field Mask",                              // Name of registered function         = can differ to "C" name: "Flat-Field Correction"
    "Array of flat-field correction mask",          // Description of function purpose     = description of this function
    0,                                              // custom user data
    0                                               // flags                               = flags
};

// Description of parameters of GetAverage function.
// This description is needed to be able to register it by MpxManager:
ExtFunctionInfo GetAverageInfo =
{
    GetAverage,                                     // Pointer to function performing an action
    3,                                              // Number of input parameters          = (DEVID ID, i32 framenumber, i32 size)
    1,                                              // Number of output parameters         = (double *data)
    {                                               // List of input parameter descriptions
        {TYPE_U32, 1, "DeviceID"},                  //                                     = first input parameter is ID of device, this device will be source of data to correct
        {TYPE_U32, 1, "FrameIndex - dummy"},        //                                     = second input parameter is index of the frame to correct (has no meaning here)
        {TYPE_U32, 1, "PixelCount"}                 //                                     = third input parameter is count of pixels - only for security reasons
    },
    {                                               // List of output parameter descriptions
        {TYPE_DOUBLE,  - 1, "Average data"}         //                                     = output: double array. Size is given by third input parameter (PixelCount)
    }, 
    "FF",                                           // Name of the plugin                  = name of this plugin "FF"
    "Average",                                      // Name of registered function         = can differ to "C" name: "Flat-Field Correction"
    "Average of all frames in the buffer",          // Description of function purpose     = description of this function
    0,                                              // custom user data
    0                                               // flags                               = flags
};

// Description of parameters of GetStdDev function.
// This description is needed to be able to register it by MpxManager:
ExtFunctionInfo GetStdDevInfo =
{
    GetStdDev,                                      // Pointer to function performing an action
    3,                                              // Number of input parameters          = (DEVID ID, i32 framenumber, i32 size)
    1,                                              // Number of output parameters         = (double *data)
    {                                               // List of input parameter descriptions
        {TYPE_U32, 1, "DeviceID"},                  //                                     = first input parameter is ID of device, this device will be source of data to correct
        {TYPE_U32, 1, "FrameIndex - dummy"},        //                                     = second input parameter is index of the frame to correct (has no meaning here)
        {TYPE_U32, 1, "PixelCount"}                 //                                     = third input parameter is count of pixels - only for security reasons
    },
    {                                               // List of output parameter descriptions
        {TYPE_DOUBLE,  - 1, "StdDev data"}          //                                     = output: double array. Size is given by third input parameter (PixelCount)
    }, 
    "FF",                                           // Name of the plugin                  = name of this plugin "FF"
    "Standard Deviation",                           // Name of registered function         = can differ to "C" name: "Flat-Field Correction"
    "Standard Deviation of all frames in the buffer",          // Description of function purpose     = description of this function
    0,                                              // custom user data
    0                                               // flags                               = flags
};

// Description of parameters of EvaluateSNR function.
// This description is needed to be able to register it by MpxManager:
ExtFunctionInfo EvaluateSNRInfo =
{
    EvaluateSNR,                                      // Pointer to function performing an action
    3,                                              // Number of input parameters          = (DEVID ID, i32 framenumber, i32 size)
    1,                                              // Number of output parameters         = (double *data)
    {                                               // List of input parameter descriptions
        {TYPE_U32, 1, "DeviceID"},                  //                                     = first input parameter is ID of device, this device will be source of data to correct
        {TYPE_U32, 1, "FrameIndex - dummy"},        //                                     = second input parameter is index of the frame to correct (has no meaning here)
        {TYPE_U32, 1, "PixelCount"}                 //                                     = third input parameter is count of pixels - only for security reasons
    },
    {                                               // List of output parameter descriptions
        {TYPE_DOUBLE,  - 1, "SNR data"}             //                                     = output: double array. Size is given by third input parameter (PixelCount)
    }, 
    "FF",                                           // Name of the plugin                  = name of this plugin "FF"
    "SNR",                                          // Name of registered function         = can differ to "C" name: "Flat-Field Correction"
    "Signal to noise ratio",                        // Description of function purpose     = description of this function
    0,                                              // custom user data
    0                                               // flags                               = flags
};

// String for path to config dir:
char configPath[350] = "";

//=========================================================================================
// Initialization:

// Use Macro PLUGIN_INIT_EXTENDED from "MpxPluginMgrAPI.h" to define:
//   - DllMain function
//   - global variable "Module" containing handle of plugin Module (HINSTANCE)
//   - global variable "mgr" pointing to table of manager functions.

PLUGIN_INIT_EXTENDED
{
    // Get number of devices and create FFmask object for each device:
    FFmaskArray.Init();

    // Register Menu function for each device:
    ITEMID menuitemID;
    for (unsigned i = 0; i < FFmaskArray.count; i++)
    {
        DEVID devID = FFmaskArray[i]->devID;
        mgr->addMpxMenuItem(devID, "FF", "Load mask", "Flat-field masks", LoadFFcoefficients, (MENUFUNCPAR)FFmaskArray[i]->devID, &menuitemID, 0);        
        mgr->addMpxMenuItem(devID, "FF", "Use sum of frames as mask", "Flat-field masks", GetFFcoefficientsFromFrames, (MENUFUNCPAR)FFmaskArray[i]->devID, &menuitemID, 0);

        // Register function performing FF correction:
        _snprintf(MakeFFCorrectionInfo.functionName, 32, "Flat-field Correction %u", devID);
        _snprintf(MakeFFCorrectionInfo.description, 128, "Flat-field correction (device %u) of frame with specified ID", devID);
        MakeFFCorrectionInfo.userData = devID;
        mgr->addFuncItem(&MakeFFCorrectionInfo);

        _snprintf(MakeFFCorrectionOfAverageInfo.functionName, 32, "Flat-field Corr. of average %u", devID);
        _snprintf(MakeFFCorrectionOfAverageInfo.description, 128, "Flat-field correction of average of all data in frame buffer of device %u", devID);
        MakeFFCorrectionOfAverageInfo.userData = devID;
        mgr->addFuncItem(&MakeFFCorrectionOfAverageInfo);
    }
    
    mgr->addFuncItem(&ShowFFMaskInfo);
    mgr->addFuncItem(&GetAverageInfo);
    mgr->addFuncItem(&GetStdDevInfo);

    mgr->addFuncItem(&EvaluateSNRInfo);

    // return value 0 indicates no error:
    return 0;
}

//=========================================================================================
// Functions to be registered by Manager:

// Let's define structure holding layout of inputs:
struct InParamStucture
{
    u32 deviceID;           // first input parameter is ID of source Device
    u32 frameIndex;         // second is index of source Frame
    u32 size;               // third is size of the frame (pixel count)
};

// Function "MakeFFCorrection" to be registered by MpxManager
// => Function is accessible for other plugins
// Function gets selected frame from selected device and correts it using flat-field appropriate coeficients.
// Resulting array will be stored in output array

int MakeFFCorrection(INTPTR userData, byte *in, byte *out)
{
    DEVID devID = (DEVID) userData;
    FRAMEID frameID = *((FRAMEID*) in);
    // Find appropriate FFmask object:
    FFmask *mask = FFmaskArray.GetMask(devID);
    if (!mask)
    {
        mgr->logMsg(PLUGIN_NAME, "Invalid device ID (devID=0x%h)", 0, devID);
        return -1;
    }
    // Make flat-field correction
    if (mask->MakeFFcorr(frameID))
        return 0;
    else
        return -1;
}


// Function making FF corection of sum of all frames.
int MakeFFCorrectionOfAverage(INTPTR userData, byte *in, byte *out){

    DEVID devID = (DEVID) userData;
    FRAMEID frameID = *((FRAMEID*) in);

    // Find appropriate FFmask object:
    FFmask *mask = FFmaskArray.GetMask(devID);
    if (!mask)
    {
        mgr->logMsg(PLUGIN_NAME, "Invalid device ID (devID=0x%h)", 0, devID);
        return -1;
    }
    // Make flat-field correction
    if (mgr->setFrameType(frameID, TYPE_DOUBLE))
        return -1;
    byte *buff = 0;
    u32 buffSize = 0;
    if (mgr->getLockedFrameBuff(frameID, &buff, &buffSize, 0))
        return -1;
    if (mask->MakeFFcorrOfAverage((double*)buff, buffSize/sizeof(double)))
        return 0;
    else
        return -1;
}

// Function will return current FF mask
int ShowFFMask(INTPTR userData, byte *in, byte *out){
    // Retype inputs to simplyfy access to each input parameter:
    InParamStucture *par = (InParamStucture*)in;

    // Find appropriate FFmask object:
    FFmask *mask = FFmaskArray.GetMask(par->deviceID);
    if (!mask){
        // not successfull: fill output by zeroes
        for (unsigned i = 0; i < par->size; i++)
          ((double*)out)[i] = 0;
        mgr->logMsg(PLUGIN_NAME, "Invalid device ID (devID=0x%h)", 0, par->deviceID);
        return -1;
    }
    if (mask->GetMask((double*)out,par->size)){
        // Success =>return
        return 0;
    }
    // not successfull: fill output by zeroes
    for (unsigned i = 0; i < par->size; i++)
        ((double*)out)[i] = 0;

    return 0;
}

int GetAverage(INTPTR userData, byte *in, byte *out){
    // Retype inputs to simplyfy access to each input parameter:
    InParamStucture *par = (InParamStucture*)in;

    // Find appropriate FFmask object:
    FFmask *mask = FFmaskArray.GetMask(par->deviceID);
    if (!mask){
        // not successfull: fill output by zeroes
        for (unsigned i = 0; i < par->size; i++)
          ((double*)out)[i] = 0;
        mgr->logMsg(PLUGIN_NAME, "Invalid device ID (devID=0x%h)", 0, par->deviceID);
        return -1;
    }
    if (mask->GetAverage((double*)out,par->size)){
        // Success =>return
        return 0;
    }
    // not successfull: fill output by zeroes
    for (unsigned i = 0; i < par->size; i++)
        ((double*)out)[i] = 0;

    return 0;
}

int GetStdDev(INTPTR userData, byte *in, byte *out){
    // Retype inputs to simplyfy access to each input parameter:
    InParamStucture *par = (InParamStucture*)in;

    // Find appropriate FFmask object:
    FFmask *mask = FFmaskArray.GetMask(par->deviceID);
    if (!mask){
        // not successfull: fill output by zeroes
        for (unsigned i = 0; i < par->size; i++)
          ((double*)out)[i] = 0;
        mgr->logMsg(PLUGIN_NAME, "Invalid device ID (devID=0x%h)", 0, par->deviceID);
        return -1;
    }
    if (mask->GetSigma((double*)out,par->size)){
        // Success =>return
        return 0;
    }
    // not successfull: fill output by zeroes
    for (unsigned i = 0; i < par->size; i++)
        ((double*)out)[i] = 0;

    return 0;
}

// Function evaluating SNR for each pixel: Signal is difference between FF mask and average data, noise is StdDev
int EvaluateSNR(INTPTR userData, byte *in, byte *out){                  
    // Retype inputs to simplyfy access to each input parameter:
    InParamStucture *par = (InParamStucture*)in;

    // Find appropriate FFmask object:
    FFmask *mask = FFmaskArray.GetMask(par->deviceID);
    if (!mask){
        // not successfull: fill output by zeroes
        for (unsigned i = 0; i < par->size; i++)
          ((double*)out)[i] = 0;
        mgr->logMsg(PLUGIN_NAME, "Invalid device ID (devID=0x%h)", 0, par->deviceID);
        return -1;
    }
    if (mask->MakeSNR((double*)out,par->size)){
        // Success =>return
        return 0;
    }
    // not successfull: fill output by zeroes
    for (unsigned i = 0; i < par->size; i++)
        ((double*)out)[i] = 0;

    return 0;
}

// Function "LoadFFcoefficients" to be registered in menu:
// - will display FileOpenDialog to select mask file
// - will load coefficients from the selected file

void LoadFFcoefficients(MENUFUNCPAR par)
{
    // par is the Device ID => have to find appropriate FFmask object:
    FFmask *mask = FFmaskArray.GetMask(par);
    if (mask)
    {
        // Prepare parameters for OpenFileDialog
        char *filtername[] = {"ASCII matrix", "*.txt", "All files", "*.*", "", ""};
        char filename[256] = "";
        strcpy(filename, mask->GetFileName());
        // Create OPENFILENAME structure:
        OPENFILENAME data;
        data.lStructSize = sizeof(OPENFILENAME);
        data.hwndOwner = NULL;
        data.hInstance = NULL;
        data.lpstrFilter = filtername[0];
        data.lpstrCustomFilter = NULL;
        data.nMaxCustFilter = NULL;
        data.nFilterIndex = 0;
        data.lpstrFile = filename;
        data.nMaxFile = 256;
        data.lpstrFileTitle = NULL;
        data.nMaxFileTitle = 0;
        data.lpstrInitialDir = NULL;
        data.lpstrTitle = "Open file with flat field coefficients";
        data.Flags = OFN_FILEMUSTEXIST;
        data.lpstrDefExt = "*.txt";
        data.lpTemplateName = NULL;

        // Show the OpenFileDialog and wait to user to choose a file:
        if (GetOpenFileName(&data))
        {
            // Try to read the file as matrix of correction coeficients
            if (!mask->LoadMask(filename))
            {
                // In case of error show an Error message box
                MessageBox(NULL, "Incompatible file format", "Error", MB_ICONERROR);
            }
        }
    }
}


// Function "GetFFcoefficientsFromFrames" to be registered in menu:
// - will set sum of currently measured frames as FF mask

void GetFFcoefficientsFromFrames(MENUFUNCPAR par)
{
    // par is the Device ID => have to find appropriate FFmask object:
    FFmask *mask = FFmaskArray.GetMask(par);
    if (mask)
    {
      if (!mask->GetMaskFromFrames())
        MessageBox(NULL,"No frames in buffer, using default coeficients","Error",MB_ICONSTOP);
    }
}

//=========================================================================================
// Methods of FFmask class - for making a standard FF correction:

// Default constructor - No initialization
FFmask::FFmask()
{
    devID = -1;
    numberOfChips = 0;
    numberOfRows = 0;
    chipboardID[0] = 0;
    ifaceName[0] = 0;
    errorMsg[0] = 0;
    mask = NULL;
    coef = NULL;
    data = NULL;
    avrg = NULL;
    stddev=NULL;
    lastFileName = NULL;
    changed = false;
    defFileName[0] = 0;
    median=1;
    frames=0;
}

// Destructor - deletes all allocated buffers, calls SaveCoefMask if needed
FFmask::~FFmask()
{
    if (changed) 
        SaveMask();
    Delete();
}

void FFmask::Delete()
{
    if (mask)
    {
        delete[] mask;
        mask = NULL;
    }
    if (coef)
    {
        delete[] coef;
        coef = NULL;
    }
    if (data)
    {
        delete[] data;
        data = NULL;
    }
    if (avrg)
    {
        delete[] avrg;
        avrg = NULL;
    }
    if (stddev)
    {
        delete[] stddev;
        stddev = NULL;
    }
    size = 0;
    if (lastFileName)
    {
        free(lastFileName);
        lastFileName = NULL;
    }
    defFileName[0] = 0;
}

bool FFmask::Init(DEVID _devID)
{
    Delete();
    if (mgr->mpxCtrlGetMedipixInfo(_devID, &numberOfChips, &numberOfRows, chipboardID, ifaceName) == 0)
    {
        devID = _devID;
        size = numberOfChips * 256 * 256;
        data = new u32[size];
        mask = new double[size];
        coef = new double[size];
        memset(data, 0, size * sizeof(u32));
        memset(mask, 0, size * sizeof(double));
        memset(coef, 0, size * sizeof(double));
        LoadMask();
        return true;
    }

    mgr->logMsg(PLUGIN_NAME, "Cannot get description of device with devID=0x%h", 0, _devID);
    return false;
}


// Will use default mask (1 in each pixel)
void FFmask::SetDefaultMask(){
  for (int i = 0; i < size; i++){
    mask[i] = 1;
    coef[i] = 1;
  }
  median=1;
}


// Twoo helper functions for mask preprocessing (median computation, normalization)
//
int doublecomp(const void *a, const void *b){
  if ( *((double*)a)==*((double*)b) )
    return(0);
  if ( *((double*)a)<*((double*)b) )
    return(-1);
  else
    return(1);
}
double DoubleSort(double* array, int length){
  qsort(array, length, sizeof(double), doublecomp);
  return(array[length/2]);
}

void FFmask::PreprocessMask(){
  memcpy(coef,mask,size*sizeof(double));
  median=DoubleSort(coef,size);
  for (int i=0;i<size;i++)
    coef[i]=(mask[i]!=0)? (median/mask[i]):0;
}

bool FFmask::GetMaskFromFrames(){
  for (int i=0;i<size;i++)
    mask[i]=0;
  int frameindex=0;
  while (mgr->mpxCtrlGetFrameDouble(devID, coef, size, frameindex)==0){
    for (int i=0;i<size;i++)
      mask[i]+=coef[i];
    frameindex++;
  }
  if (!frameindex){
    SetDefaultMask();   
    mgr->logMsg(PLUGIN_NAME, "GetMaskFromFrames: Error getting frames => setting default FF mask.", 0, frameindex);
    return(false);
  }
  PreprocessMask();
  changed = true;
  return(true);
}


bool FFmask::LoadMask(char *fileName)
{
    SetFileName(NULL);
    changed = false;
    if (!coef)
        return false;

    if (!fileName)
        fileName = GetDefFileName();

    FILE *F = fopen(fileName, "r");
    if (!F)
    {
        mgr->logMsg(PLUGIN_NAME, "Error loading FF coeficients. Cannot open file: \r%s", 0, fileName);
        return false;
    }

    // Loading mask data:
    for (int i = 0; i < size; i++)
    {
        if (fscanf(F, "%lg", &mask[i]) != 1)
        {
            SetDefaultMask();   
            fclose(F);
            mgr->logMsg(PLUGIN_NAME, "Error loading FF mask => setting default FF mask. Can't read from: \r%s", 0, fileName);
            return false;
        }
    }
    
    // Preprocessing: Inverting and normalising by median
    PreprocessMask();

    fclose(F);
    SetFileName(fileName);
    changed = true;
    return true;
}

// Saves coeficients to ASCII file. If filename is NULL generates default.
bool FFmask::SaveMask(char *fileName)
{
    if (!coef)
        return false;

    if (!fileName)
        fileName = GetDefFileName();

    FILE *F = fopen(fileName, "w");
    if (!F)
    {
        mgr->logMsg(PLUGIN_NAME, "Error writting FF coeficients. Cannot open file: \r%s", 0, fileName);
        return false;
    }
    int columns = size / (numberOfRows *256);
    for (int i = 0; i < size; i++)
        if (fprintf(F, ((i + 1) % columns) ? "%lg\t" : "%lg\n", mask[i]) < 0)
            mgr->logMsg(PLUGIN_NAME, "Error writting FF mask to: \r%s", 0, fileName);
    fclose(F);
    return true;
}

// Returns default filename
char *FFmask::GetDefFileName()
{
    if (!configPath[0])
    {
        // erase defFileName as path is unknown yet:
        defFileName[0]=0;
        
        // acquire path to this plugin dll:
        if (!GetModuleFileName((HMODULE)Module, configPath, 350))
            configPath[0] = 0;

        // find first backslash from the right side and cut it off
        char *beg = strrchr(configPath, '\\');
        if (!beg)
            beg = configPath;
        *beg = 0;
        
        // find second backslash from the right side and copy the config/
        beg = strrchr(configPath, '\\');
        if (!beg)
            beg = configPath;
        else
            strcpy(beg, "\\configs\\");
    }

    if (!defFileName[0])
        sprintf(defFileName, "%s%s_%s_ff.txt", configPath, ifaceName, chipboardID);

    return defFileName;
}


bool FFmask::MakeFFcorr(FRAMEID frameID)
{
    if (!coef)
    {
        mgr->logMsg(PLUGIN_NAME, "Error making FF corection - Not initialized", 0);
        return false;
    }
    if (mgr->setFrameType(frameID, TYPE_DOUBLE))
        return false;
    byte *buff = 0;
    u32 buffSize = 0;
    if (mgr->getLockedFrameBuff(frameID, &buff, &buffSize, 0))
        return false;
    if (buffSize/sizeof(double) != size)
    {
        mgr->logMsg(PLUGIN_NAME, "Wrong frame dimension", 0);
        mgr->unlockFrame(frameID);
        return false;
    }
    double *data = (double *) buff;
    for (int i = 0; i < size; i++)
        data[i] *= coef[i];
    mgr->unlockFrame(frameID);
    return true;
}

// TakeAverage call GetFrame32 function from MpxManager in a loop and takes average of all measured frames and their
// Standard deviation. These arrays are stored in member arrays avrg and stddev. Member frames will be also filled.
bool FFmask::TakeAverage(){
  if (!avrg)
    avrg = new double[size];
  for (int i=0;i<size;i++)
    avrg[i]=0;
  frames=0;
  u32 filled = 0;
  mgr->mpxCtrlGetAcqInfo(devID, 0, 0, 0, &filled);
  for (u32 k = 0; k < filled; k++)
  {
      if (mgr->mpxCtrlGetFrame32(devID, data, size, frames))
          break;
      for (int i=0;i<size;i++)
          avrg[i]+=data[i];
      frames++;
  }
  if (!frames){
    mgr->logMsg(PLUGIN_NAME, "Error making average - No data in framebuffer", 0);
    return(false);
  }
  if (frames>1)
    for (int i = 0; i < size; i++)
      avrg[i] /= frames;
  return(true);
}


// TakeSigma call GetFrame32 function from MpxManager in a loop and takes stddev of all measured frames. 
// Result is stored in member array stddev. Member frames will be also filled.
bool FFmask::TakeAverageAndSigma(){
  if ((!TakeAverage())||(frames<2)){
    mgr->logMsg(PLUGIN_NAME, "Error making standard deviation - Not enough data in framebuffer", 0);
    return(false);
  }
  if (!stddev)
    stddev = new double[size];
  for (int i=0;i<size;i++)
    stddev[i]=0;
  double diff;
  for (int j=0; j<frames;j++){
    if (mgr->mpxCtrlGetFrame32(devID, data, size, j)!=0){
      mgr->logMsg(PLUGIN_NAME, "Error making standard deviation - Number of frames has been changed", 0);
      return(false);
    }
    for (int i=0;i<size;i++){
      diff=avrg[i]-data[i];
      stddev[i]+=diff*diff;
    }
  }
  for (int i = 0; i < size; i++)
    stddev[i]=sqrt(stddev[i]/(frames-1));
  return(true);
}


bool FFmask::MakeFFcorrOfAverage(double *FFData, int dsize){
  if ((!data) || (!coef) || (dsize != size)){
    mgr->logMsg(PLUGIN_NAME, "Error making FF corection - Not initialized", 0);
    return false;
  }
  if (!TakeAverage())
    return(false);
  if (FFData)
    for (int i = 0; i < size; i++)
      FFData[i] = avrg[i] *coef[i];
  return true;
}

bool FFmask::GetMask(double *array, int dsize){
  if ((!mask) || (dsize != size)){
    mgr->logMsg(PLUGIN_NAME, "Error reading FF corection mask - Not initialized", 0);
    return false;
  }
  for (int i = 0; i < size; i++)
    array[i] = mask[i];
  return true;
}

bool FFmask::GetAverage(double *array, int dsize){
  if ((!data) || (!coef) || (dsize != size)){
    mgr->logMsg(PLUGIN_NAME, "Error making FF corection - Not initialized", 0);
    return false;
  }
  if (!TakeAverage())
    return(false);
  if (array)
    for (int i = 0; i < size; i++)
      array[i] = avrg[i];
  return true;
}

bool FFmask::GetSigma(double *array, int dsize){
  if ((!data) || (!coef) || (dsize != size)){
    mgr->logMsg(PLUGIN_NAME, "Error making FF corection - Not initialized", 0);
    return false;
  }
  if (!TakeAverageAndSigma())
    return(false);
  if (array)
    for (int i = 0; i < size; i++)
      array[i] = stddev[i];
  return true;
}
    
// MakeSNR computes SNR of each pixel. Signal is diference between average and mask. Noise is StdDev.
bool FFmask::MakeSNR(double *array, int dsize){
  if ((!data) || (!coef) || (dsize != size)){
    mgr->logMsg(PLUGIN_NAME, "Error making FF corection - Not initialized", 0);
    return false;
  }
  if (!TakeAverageAndSigma())
    return(false);
  if (array)
    for (int i = 0; i < size; i++)
      array[i] = (stddev[i]>0)? (fabs(avrg[i]-mask[i])/stddev[i]):0;
  return true;
}


//=========================================================================================
// Methods of FFmaskArrayType class - array of FFmask objects:

// Initialization of the array
void FFmaskArrayType::Init()
{
    // Delete current array, save configuration:
    Delete();

    DEVID devID;
    // Get number of devices:
    if (mgr->mpxCtrlGetFirstMpx(&devID, (int*) &count) != 0)
    {
        mgr->logMsg(PLUGIN_NAME, "Cannot get number of devices", 0);
        return ;
    }
    // Create FFmask object for each device:
    if (count)
    {
        mask = new FFmask[count];
        unsigned i = 0;
        mask[i++].Init(devID);
        while (!mgr->mpxCtrlGetNextMpx(&devID) && (i < count))
            mask[i++].Init(devID);
    }
}

// Deletes all FFmask Objects
void FFmaskArrayType::Delete()
{
    if (mask)
    {
        delete []mask;
        mask = NULL;
    }
    count = 0;
}

// GetMask will search for FFmask object with given devID:
// Not very effective but not many items in Mask array are expected
FFmask *FFmaskArrayType::GetMask(DEVID devID)
{
    for (unsigned i = 0; i < count; i++)
        if (mask[i].devID == devID)
            return  &mask[i];

    return NULL;
}

